import logging
from rich.logging import RichHandler
from rich.console import Console
from rich.progress import track

import argparse
import os, pathlib
from pickle import Pickler, Unpickler

from bloodhound.ad.utils import ADUtils

from collections import defaultdict
from requests.structures import CaseInsensitiveDict

import datetime
from enum import Enum

class ADExplorerSnapshot(object):
    OutputMode = Enum('OutputMode', ['BOFHound', 'BloodHound', 'Objects'])

    def __init__(self, snapfile, outputfolder, console=None, snapshot_parser=None):
        self.console = console or setup_logging()
        self.output = outputfolder

        if not snapshot_parser:
            from adexpsnapshot.parser.classes import Snapshot
            snapshot_parser = Snapshot

        self.snapfile = snapfile
        self.snap = snapshot_parser(snapfile)

        self.snap.parseHeader()

        filetimeiso = datetime.datetime.fromtimestamp(self.snap.header.filetimeUnix).isoformat()
        logging.info(f'Server: {self.snap.header.server}')
        logging.info(f'Time of snapshot: {filetimeiso}')
        logging.info('Metadata offset: 0x{:x}'.format(self.snap.header.metadataOffset))
        logging.info(f'Object count: {self.snap.header.numObjects}')

        self.snap.parseProperties()
        self.snap.parseClasses()
        self.snap.parseObjectOffsets()

        self.sidcache = {}
        self.dncache = CaseInsensitiveDict()
        self.computersidcache = CaseInsensitiveDict()
        self.domains = CaseInsensitiveDict()
        self.objecttype_guid_map = CaseInsensitiveDict()
        self.domaincontrollers = []
        self.certtemplates = defaultdict(set)

        self.rootdomain = None
        self.domain_dn = None
        self.config_dn = None
        self.schema_dn = None
        self.forest_dn = None

    def preprocess(self, cache=False):
        if cache:
            cacheFileName = str(pathlib.Path(self.snapfile.name).with_suffix('.cache'))

            dico = None
            try:
                dico = Unpickler(open(cacheFileName, "rb")).load()
            except (OSError, IOError, EOFError) as e:
                pass

            if dico and dico.get('unixtime') == self.snap.header.filetimeUnix:
                self.objecttype_guid_map = dico['guidmap']
                self.sidcache = dico['sidcache']
                self.dncache = dico['dncache']
                self.computersidcache = dico['computersidcache']
                self.domains = dico['domains']
                self.domaincontrollers = dico['domaincontrollers']
                self.certtemplates = dico['certtemplates']

                self.rootdomain = dico['rootdomain']
                self.domain_dn = dico['domain_dn']
                self.config_dn = dico['config_dn']
                self.schema_dn = dico['schema_dn']
                self.forest_dn = dico['forest_dn']

                self.console.print(f"[green]✓[/green] Using cached data for preprocessing")
                return

        for k, cl in self.snap.classes.items():
            self.objecttype_guid_map[k] = str(cl.schemaIDGUID)

        for k, idx in self.snap.propertyDict.items():
            self.objecttype_guid_map[k] = str(self.snap.properties[idx].schemaIDGUID)

        for idx, obj in track(enumerate(self.snap.objects), description="Preprocessing", total=self.snap.header.numObjects):
            # create sid cache
            objectSid = ADUtils.get_entry_property(obj, 'objectSid')
            if objectSid:
                self.sidcache[str(objectSid)] = idx

            # create dn cache
            distinguishedName = ADUtils.get_entry_property(obj, 'distinguishedName')
            if distinguishedName:
                self.dncache[str(distinguishedName)] = idx

            # get domains
            # Domain objects have 'domain' class AND objectSid attribute
            # DNS zones have 'domain' class but no objectSid
            if 'domain' in obj.classes and objectSid:
                if self.rootdomain is not None:  # is it possible to find multiple?
                    logging.warning("Multiple domains in snapshot(?)")
                else:
                    self.rootdomain = str(distinguishedName) # keeping for backward compat
                    self.domain_dn = str(distinguishedName)
                    self.domains[str(distinguishedName)] = idx

                    objectCategory = ADUtils.get_entry_property(obj, 'objectCategory')
                    try:
                        _, self.schema_dn = objectCategory.split(',', 1)
                        _, self.config_dn = self.schema_dn.split(',', 1)
                        _, self.forest_dn = self.config_dn.split(',', 1)
                    except (ValueError, AttributeError):
                        logging.warning("Failed to parse schema, config, or forest DN from objectCategory")
                        pass

            # get forest domains
            if 'crossref' in obj.classes:
                if ADUtils.get_entry_property(obj, 'systemFlags', 0) & 2 == 2:
                    ncname = ADUtils.get_entry_property(obj, 'nCName')
                    if ncname and ncname not in self.domains:
                        self.domains[str(ncname)] = idx

            # get computers
            if ADUtils.get_entry_property(obj, 'sAMAccountType', -1) == 805306369:
                dnshostname = ADUtils.get_entry_property(obj, 'dNSHostname')
                if dnshostname:
                    self.computersidcache[str(dnshostname)] = str(objectSid)

            # get all cert templates
            if 'pkienrollmentservice' in obj.classes:
                name = str(ADUtils.get_entry_property(obj, 'name'))
                if ADUtils.get_entry_property(obj, 'certificateTemplates'):
                    templates = ADUtils.get_entry_property(obj, 'certificateTemplates')
                    for template in templates:
                        self.certtemplates[str(template)].add(name)

            # get dcs
            if ADUtils.get_entry_property(obj, 'userAccountControl', 0) & 0x2000 == 0x2000:
                self.domaincontrollers.append(idx)

        if cache:
            dico = {
                'guidmap': self.objecttype_guid_map,
                'sidcache': self.sidcache,
                'dncache': self.dncache,
                'computersidcache': self.computersidcache,
                'domains': self.domains,
                'domaincontrollers': self.domaincontrollers,
                'rootdomain': self.rootdomain,
                'domain_dn': self.domain_dn,
                'config_dn': self.config_dn,
                'schema_dn': self.schema_dn,
                'forest_dn': self.forest_dn,
                'certtemplates': self.certtemplates,
                'unixtime': self.snap.header.filetimeUnix
            }
            with open(cacheFileName, "wb") as f:
                Pickler(f).dump(dico)
            self.console.print(f"[green]✓[/green] Cache saved to {cacheFileName}")
        
        self.console.print(f"[green]✓[/green] Preprocessing complete")

    def preprocessCached(self):
        """Alias for preprocess() for backward compatibility with scripts."""
        self.preprocess(cache=True)

    def outputObjects(self):
        """Output all objects to NDJSON file."""
        from adexpsnapshot.ouput.objects import ObjectsOutput
        handler = ObjectsOutput(self.snap, self.output, self.console)
        handler.process()

    def outputBOFHound(self):
        """Output objects in BOFHound log format."""
        from adexpsnapshot.ouput.bofhound import BOFHoundOutput
        handler = BOFHoundOutput(self.snap, self.output, self.console)
        handler.process()

    def outputBloodHound(self):
        """Output data in BloodHound JSON format."""

        logging.warning("The BloodHound mode gives incomplete output. Please use the BOFHound mode instead.")

        self.preprocess(cache=True)
        from adexpsnapshot.ouput.bloodhound import BloodHoundOutput
        handler = BloodHoundOutput(
            self.snap, self.output, self.console,
            self.sidcache, self.dncache, self.computersidcache, self.domains,
            self.objecttype_guid_map, self.domaincontrollers, self.rootdomain, self.certtemplates
        )
        handler.process()


def setup_logging(level=logging.INFO):
    console = Console()
    rich_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=False,
        rich_tracebacks=True,
        markup=True
    )
    
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[rich_handler]
    )
    
    return console

def mode_output(ades, args):
    """Handle output mode operations (BOFHound, BloodHound, Objects)."""
    outputmode = ADExplorerSnapshot.OutputMode[args.mode]
    if outputmode == ADExplorerSnapshot.OutputMode.BloodHound:
        ades.outputBloodHound()
    elif outputmode == ADExplorerSnapshot.OutputMode.Objects:
        ades.outputObjects()
    elif outputmode == ADExplorerSnapshot.OutputMode.BOFHound:
        ades.outputBOFHound()

def mode_enrich(ades):
    """Handle enrichment mode - reconstruct missing treeview metadata."""
    from adexpsnapshot.enrich import enrich_snapshot
    res = enrich_snapshot(ades)
    if res:
        ades.console.print(f"[green]✓[/green] Enrichment complete")
    else:
        ades.console.print(f"[red]✗[/red] Enrichment skipped or failed")

def main():
    console = setup_logging()

    parser = argparse.ArgumentParser(add_help=True, description='AD Explorer snapshot ingestor for BloodHound', formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('snapshot', type=argparse.FileType('rb'), help="Path to the snapshot .dat file.")
    parser.add_argument('-o', '--output', required=False, type=pathlib.Path, help="Path to an output folder. Folder will be created if it doesn't exist. Defaults to the current directory.", default=".")
    parser.add_argument('-m', '--mode', required=False, help="The output mode to use. Defaults to BOFHound output mode, which can then be used with BOFHound. Can also directly output to BloodHound JSON output files (with limitations). In Objects mode all objects with all attributes are outputted to NDJSON.", choices=ADExplorerSnapshot.OutputMode.__members__, default='BOFHound')
    parser.add_argument('-e', '--enrich', action='store_true', help="Reconstruct treeview metadata if missing and save to enriched file.")

    args = parser.parse_args()

    if not os.path.exists(args.output):
        try:
            os.mkdir(args.output)
        except:
            logging.error(f"Unable to create output directory '{args.output}'.")
            return
    
    if not os.path.isdir(args.output):
        logging.error(f"Path '{args.output}' does not exist or is not a folder.")
        return
    
    ades = ADExplorerSnapshot(args.snapshot, args.output, console=console)
    
    if args.enrich:
        mode_enrich(ades)
    else:
        mode_output(ades, args)

if __name__ == '__main__':
    main()



from dissect.cstruct import cstruct

treeview_structure = cstruct()
treeview_structure.load("""
    // Treeview header - supports variable number of naming contexts
    struct TreeviewHeader {
        uint64 magic;                   // 0xFFFFFFFFFFFFFFFE
        uint32 num_NCs;                 // Number of naming contexts (usually 3: Domain, Configuration, Schema)
        uint32 reserved;                // 0
        uint32 section_offsets[num_NCs]; // Byte offsets to each NC section (relative to start of treeview metadata)
    };
    
    // Parent node - complete structure with variable-length arrays
    // All nodes in the treeview are ParentNodes
    // (Leaf children are stored inline in their parent's inline_child_offsets)
    struct ParentNode {
        uint64 objectOffset;
        uint32 num_children_with_children;    // Number of children that have their own ParentNode entries
        uint32 num_children_without_children; // Number of leaf children (stored inline below)
        uint32 child_offsets[num_children_with_children];           // Relative byte offsets to child ParentNode entries
        uint64 inline_child_offsets[num_children_without_children]; // Direct objectOffsets for leaf children (no standalone entry)
    };
    
""", compiled=True)




"""
Create synthetic AD objects for missing intermediate containers.

When building treeview, some DNs may reference parent containers that don't
exist as actual objects in the snapshot. This module creates minimal synthetic
objects for these missing parents.
"""

import struct
from io import BytesIO


def create_synthetic_object(dn, snap):
    """
    Create a minimal synthetic AD object for a missing container DN.
    
    Returns bytes of the object structure.
    
    Structure:
    - objSize (uint32)
    - tableSize (uint32) 
    - mappingTable entries (8 bytes each)
    - attribute data
    """
    
    # Find property index for distinguishedName
    dn_prop_idx = snap.propertyDict.get('distinguishedName')
    
    if dn_prop_idx is None:
        raise ValueError("Cannot create synthetic object: missing distinguishedName property")
    
    # Only one attribute: distinguishedName
    table_size = 1
    
    # Attribute data starts after: objSize(4) + tableSize(4) + mappingTable(1 * 8) = 16
    attr_data_start = 8 + (table_size * 8)
    
    # Build attribute data
    attr_buf = BytesIO()
    
    # Attribute: distinguishedName
    # Format: [numValues][offset_array][string_data]
    dn_attr_offset = attr_data_start
    dn_encoded = dn.encode('utf-16le') + b'\x00\x00'
    attr_buf.write(struct.pack("<I", 1))  # num values
    attr_buf.write(struct.pack("<I", 8))  # offset within attribute (after numValues + offset = 8)
    attr_buf.write(dn_encoded)
    
    # Build mapping table
    mapping_table = [
        (dn_prop_idx, dn_attr_offset)
    ]
    
    # Calculate total size
    attr_data = attr_buf.getvalue()
    obj_size = 8 + (table_size * 8) + len(attr_data)
    
    # Build complete object
    obj_buf = BytesIO()
    obj_buf.write(struct.pack("<I", obj_size))
    obj_buf.write(struct.pack("<I", table_size))
    
    for attr_idx, attr_offset in mapping_table:
        obj_buf.write(struct.pack("<I", attr_idx))  # attrIndex (unsigned)
        obj_buf.write(struct.pack("<i", attr_offset))  # attrOffset (signed, can be negative)
    
    obj_buf.write(attr_data)
    
    return obj_buf.getvalue()


def create_synthetic_objects_data(missing_dns, snap, start_offset):
    """
    Create synthetic AD objects for missing intermediate containers.
    
    Args:
        missing_dns: Set of DN strings for missing parents
        snap: Snapshot object (for property indices)
        start_offset: File offset where synthetic objects will be written
    
    Returns:
        (synthetic_offsets, synthetic_data_bytes)
        synthetic_offsets: DN -> obj_offset (int)
        synthetic_data_bytes: concatenated binary data for all synthetic objects
    """
    if not missing_dns:
        return {}, b''
    
    from io import BytesIO
    
    synthetic_objects = {}
    all_data = BytesIO()
    current_offset = start_offset
    
    # Sort by DN length to create parents before children
    for dn in sorted(missing_dns, key=len):
        obj_data = create_synthetic_object(dn, snap)
        
        synthetic_objects[dn] = current_offset
        
        all_data.write(obj_data)
        current_offset += len(obj_data)
    
    return synthetic_objects, all_data.getvalue()




"""BOFHound output mode - outputs objects in BOFHound log format."""
import codecs
import base64
import datetime
import logging
import os
import queue
import threading
from rich.console import Console
from rich.progress import track


class BOFHoundEncoder:
    """Encoder for BOFHound log format."""
    
    timestamp_attributes = ['whenCreated', 'whenChanged', 'dSCorePropagationData']

    def encode(self, obj):
        if obj is None:
            return ""
        elif isinstance(obj, dict):
            return self.encode_dict(obj)
        elif isinstance(obj, list):
            return ", ".join(self.encode(item) for item in obj)
        elif isinstance(obj, int):
            return str(obj if obj < 0x80000000 else obj - 0x100000000)
        elif isinstance(obj, bytes):
            return base64.b64encode(obj).decode("ascii")
        elif isinstance(obj, (str, float, bool)):
            return str(obj)
        else:
            raise Exception(f"BOFHoundEncoder does not support objects of type {type(obj)}")

    def encode_dict(self, obj):
        lines = []
        for key in sorted(obj.keys()):
            if key in BOFHoundEncoder.timestamp_attributes:
                encoded = self.encode_timestamp(obj[key], key)
            else:
                encoded = self.encode(obj[key])
            lines.append(f"{key}: {encoded}")
        if len(lines):
            lines.append("")
        return "\n".join(lines)

    def encode_timestamp(self, value, attr=''):
        try:
            if isinstance(value, list):
                if len(value) == 0:
                    return "0"
                value = value[0] & 0xFFFFFFFF
            return datetime.datetime.fromtimestamp(value, datetime.UTC).strftime('%Y%m%d%H%M%S.0Z')
        except:
            logging.warning(f"Failed to parse timestamp for attribute {attr}")
            return "0"


class BOFHoundOutput:
    """Handles BOFHound-specific output processing."""
    
    def __init__(self, snapshot, output_folder, console: Console):
        self.snap = snapshot
        self.output = output_folder
        self.console = console
        self.outputfile = f"{snapshot.header.server}_{snapshot.header.filetimeUnix}_bofhound.log"
        self.encoder = BOFHoundEncoder()

    def process(self):
        """Process all objects and output to BOFHound log file."""
        def write_worker(result_q, filename):
            try:
                fh_out = codecs.open(filename, 'w', 'utf-8')
            except:
                logging.warning('Could not write file: %s', filename)
                result_q.task_done()
                return

            wroteOnce = False
            while True:
                data = result_q.get()

                if data is None:
                    break

                if not wroteOnce:
                    wroteOnce = True
                else:
                    fh_out.write('--------------------\n')

                try:
                    encoded_member = self.encoder.encode(data)
                    fh_out.write(encoded_member)
                except TypeError:
                    logging.error('Data error {0}, could not convert data to BOFHound log'.format(repr(data)))
                result_q.task_done()

            fh_out.close()
            result_q.task_done()

        wq = queue.Queue()
        results_worker = threading.Thread(target=write_worker, args=(wq, os.path.join(self.output, self.outputfile)))
        results_worker.daemon = True
        results_worker.start()

        for idx, obj in track(enumerate(self.snap.objects), description="Dumping objects", total=self.snap.header.numObjects):
            wq.put((dict(obj.attributes.data)))

        wq.put(None)
        wq.join()

        self.console.print(f"[green]✓[/green] Output written to {self.outputfile}")



from adexpsnapshot.parser.structure import structure
from bloodhound.ad.utils import ADUtils
from bloodhound.enumeration.acls import LdapSid
from requests.structures import CaseInsensitiveDict

import struct
from collections import UserDict
import logging

import functools
import uuid
from io import BytesIO
import datetime, calendar

ADSTYPE_INVALID = 0
ADSTYPE_DN_STRING = 1
ADSTYPE_CASE_EXACT_STRING = 2
ADSTYPE_CASE_IGNORE_STRING = 3
ADSTYPE_PRINTABLE_STRING = 4
ADSTYPE_NUMERIC_STRING = 5
ADSTYPE_BOOLEAN = 6
ADSTYPE_INTEGER = 7
ADSTYPE_OCTET_STRING = 8
ADSTYPE_UTC_TIME = 9
ADSTYPE_LARGE_INTEGER = 10
ADSTYPE_PROV_SPECIFIC = 11
ADSTYPE_OBJECT_CLASS = 12
ADSTYPE_CASEIGNORE_LIST = 13
ADSTYPE_OCTET_LIST = 14
ADSTYPE_PATH = 15
ADSTYPE_POSTALADDRESS = 16
ADSTYPE_TIMESTAMP = 17
ADSTYPE_BACKLINK = 18
ADSTYPE_TYPEDNAME = 19
ADSTYPE_HOLD = 20
ADSTYPE_NETADDRESS = 21
ADSTYPE_REPLICAPOINTER = 22
ADSTYPE_FAXNUMBER = 23
ADSTYPE_EMAIL = 24
ADSTYPE_NT_SECURITY_DESCRIPTOR = 25
ADSTYPE_UNKNOWN = 26
ADSTYPE_DN_WITH_BINARY = 27
ADSTYPE_DN_WITH_STRING = 28

class WrapStruct(object):
    def __init__(self, snap, in_obj=None):
        self.snap = snap
        self.fh = snap.fh

        if in_obj:
            self._data = in_obj
        else:
            self._data = getattr(structure, type(self).__name__)(self.fh)

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):
            raise AttributeError

        return getattr(self._data, attr)

class SystemTime(WrapStruct):
    def __init__(self, snap=None, in_obj=None):
        super().__init__(snap, in_obj)

        d = datetime.datetime(self.wYear, self.wMonth, self.wDay, self.wHour, self.wMinute, self.wSecond)
        self.unixtimestamp = calendar.timegm(d.timetuple())

    def __repr__(self):
        return str(self.unixtimestamp)

# could probably use a rewrite, but it allows us to refer to attributes dynamically,
# meaning they will be retrieved/processed when necessary 
class AttributeDict(UserDict):
    def __init__(self, obj, raw):
        self.obj = obj
        self.snap = obj.snap
        self.fh = obj.fh
        self.raw = raw

        self._dico = CaseInsensitiveDict()
        
    def __getitem__(self,  key):
        ret = self.getAttribute(key, raw=self.raw)
        if key.lower() == 'name': # hacked in to make resolve_ad_entry function work 
            return ret[0]
        return ret

    @property
    def data(self):
        if len(self._dico) > 0:
            return self._dico

        for entry in self.obj.mappingTable:
            prop = self.snap.properties[entry.attrIndex]
            self._dico[prop.propName] = self.processAttribute(prop, entry.attrOffset, self.raw)

        return self._dico

    def getAttribute(self, attrName, raw=False):
        attrIndex = self.snap.propertyDict[attrName]
        prop = self.snap.properties[attrIndex]

        for entry in self.obj.mappingTable:
            if entry.attrIndex == attrIndex:
                return self.processAttribute(prop, entry.attrOffset, raw)

        raise KeyError

    def processAttribute(self, prop, attrOffset, raw):

        attrName = prop.propName.lower()
        attrType = prop.adsType

        # at the offset at which the attribute is stored, 
        #  - the first quad indicates how many elements are in the attribute (attributes can be multi-valued), 
        #  - the bytes after depend on what sort of information is stored (e.g. for DN_STRING, the quads after are the offsets at which the element values are stored)

        fileAttrOffset = self.obj.fileOffset + attrOffset
        self.fh.seek(fileAttrOffset)
        numValues = structure.uint32(self.fh)

        values = []

        # https://docs.microsoft.com/en-us/windows/win32/api/iads/ns-iads-adsvalue
        # https://docs.microsoft.com/en-us/windows/win32/adsi/adsi-simple-data-types

        if attrType in [ADSTYPE_DN_STRING, ADSTYPE_CASE_EXACT_STRING, ADSTYPE_CASE_IGNORE_STRING, ADSTYPE_PRINTABLE_STRING, ADSTYPE_NUMERIC_STRING, ADSTYPE_OBJECT_CLASS]:
            offsets = structure.uint32[numValues](self.fh)

            for v in range(numValues):
                self.fh.seek(fileAttrOffset + offsets[v]) # this can also be a negative offset, e.g. referencing data in a previous object
                val = structure.wchar[None](self.fh)
                values.append(val)

        elif attrType == ADSTYPE_OCTET_STRING:
            lengths = structure.uint32[numValues](self.fh)

            for v in range(numValues):
                octetStr = structure.char[lengths[v]](self.fh)
                val = octetStr

                if not raw:
                    if len(octetStr) == 16 and attrName.endswith("guid"):
                        val = str(uuid.UUID(bytes_le=octetStr))
                    elif attrName == 'objectsid' or attrName == 'securityidentifier':
                        val = str(LdapSid(BytesIO(octetStr)))

                values.append(val)

        elif attrType == ADSTYPE_BOOLEAN:
            for v in range(numValues):
                val = bool(structure.uint32(self.fh))
                values.append(val)

        elif attrType == ADSTYPE_INTEGER:

            for v in range(numValues):
                # defined as DWORD, so reading as uint32 (unsigned)
                val = structure.uint32(self.fh)
                values.append(val)

        elif attrType == ADSTYPE_LARGE_INTEGER:

            for v in range(numValues):
                # defined as LARGE_INTEGER, interestingly this is an int64 (signed) according to MS docs
                val = structure.int64(self.fh)
                values.append(val)

        elif attrType == ADSTYPE_UTC_TIME: # note that date/times can also be returned as Interval type instead (ADSTYPE_LARGE_INTEGER) - actual time units depend on which attribute is using it

            for v in range(numValues):
                systime = SystemTime(self.snap)
                val = systime.unixtimestamp
                values.append(val)

        elif attrType == ADSTYPE_NT_SECURITY_DESCRIPTOR:

            for v in range(numValues):
                lenDescriptorBytes = structure.uint32(self.fh)
                descriptorBytes = self.fh.read(lenDescriptorBytes)
                values.append(descriptorBytes)

        else:
            logging.warning("Unhandled adsType: %s -> %d" % (attrName, attrType))

        return values

class Object(WrapStruct):
    def __init__(self, snap=None, in_obj=None):
        super().__init__(snap, in_obj)

        self.fileOffset = self.fh.tell() - 4 - 4 - (self.tableSize * 8)
        self.fh.seek(self.fileOffset + self.objSize) # move file pointer to the next object

        self.attributes = AttributeDict(self, raw=False)
        self.raw_attributes = AttributeDict(self, raw=True)

    @functools.cached_property
    def classes(self):
        try:
            return list(map(str.casefold, self.attributes.get('objectClass', [])))
        except Exception as e:
            print("Error in resolving objectClass:")
            print(e)
        return []

    @functools.cached_property
    def category(self):
        catDN = self.attributes.get('objectCategory', None)
        if catDN is None:
            return None

        catDN = catDN[0]
        catObj = self.snap.classes.get(catDN, None)
        if catObj:
            return catObj.className.lower()
        else:
            return None

    # for easy compatibility with the bloodhound lib
    def __getitem__(self, key):
        if key == "attributes":
            return self.attributes
        elif key == "raw_attributes":
            return self.raw_attributes
        else:
            return None

class Property(WrapStruct):
    def __init__(self, snap=None, in_obj=None):
        super().__init__(snap, in_obj)

        self.propName = self.propName.rstrip('\x00')
        self.DN = self.DN.rstrip('\x00')
        self.schemaIDGUID = uuid.UUID(bytes_le=self.schemaIDGUID)

class Class(WrapStruct):
    def __init__(self, snap=None, in_obj=None):
        super().__init__(snap, in_obj)

        self.className = self.className.rstrip('\x00')
        self.DN = self.DN.rstrip('\x00')
        self.schemaIDGUID = uuid.UUID(bytes_le=self.schemaIDGUID)

class Header(WrapStruct):
    def __init__(self, snap, in_obj=None):
        super().__init__(snap, in_obj)

        self.server = self.server.rstrip('\x00')
        self.filetimeUnix = ADUtils.win_timestamp_to_unix(self.filetime)

class Snapshot(object):
    def __init__(self, fh):
        self.fh = fh
        self.objectOffsets = {}

        # the order in which we're parsing matters, due to the file handle's position
        # typically, you would call as follows:

        # self.parseHeader()
        # self.parseObjectOffsets()
        # self.parseProperties()
        # self.parseClasses()
        # self.parseRights()

    def parseHeader(self):
        self.fh.seek(0)
        self.header = Header(self)

    def parseObjectOffsetsGenerator(self):
        self.fh.seek(0x43e)

        # we are only keeping offsets at this stage, as some databases grow very big
        self.objectOffsets = []
        for i in range(self.header.numObjects):
            pos = self.fh.tell()
            objSize = struct.unpack("<I", self.fh.read(4))[0]

            self.objectOffsets.append(pos) 
            # using struct instead of dissect.cstruct here for speed
            #self.objectOffsets.append(Object(self).fileOffset)
            self.fh.seek(pos+objSize)
            yield i

    def parseObjectOffsets(self):
        list(self.parseObjectOffsetsGenerator())
        logging.info(f"Parsed {len(self.objectOffsets)} object offsets")

    def getObject(self, i):
        self.fh.seek(self.objectOffsets[i])
        return Object(self)

    def getObjects(self):
        i = 0
        while i < self.header.numObjects:
            yield self.getObject(i)
            i += 1

    objects = property(getObjects)

    def parseProperties(self):
        self.fh.seek(self.header.metadataOffset)

        properties_with_header = structure.Properties(self.fh)
        self.properties = []
        self.propertyDict = CaseInsensitiveDict()

        for idx, p in enumerate(properties_with_header.properties):
            prop = Property(self, in_obj=p)
            self.properties.append(prop)

            # abuse our dict for both DNs and the display name / cn
            self.propertyDict[prop.propName] = idx
            self.propertyDict[prop.DN] = idx
            self.propertyDict[prop.DN.split(',')[0].split('=')[1]] = idx

        logging.info(f"Parsed {len(self.properties)} properties")

    def parseClasses(self):
        classes_with_header = structure.Classes(self.fh)
        self.classes = CaseInsensitiveDict()
        for c in classes_with_header.classes:
            cl = Class(self, in_obj=c)

            # abuse our dict for both DNs and the display name / cn
            self.classes[cl.className] = cl
            self.classes[cl.DN] = cl
            self.classes[cl.DN.split(',')[0].split('=')[1]] = cl

        logging.info(f"Parsed {len(self.classes)} classes")

    def parseRights(self):
        rights_with_header = structure.Rights(self.fh)
        self.rights = rights_with_header.rights

        logging.info(f"Parsed {len(self.rights)} rights")



"""BloodHound output mode - outputs data in BloodHound JSON format."""
import logging
import os
import queue
import threading
import functools
from typing import List
from rich.console import Console
from rich.progress import track

from bloodhound.ad.utils import ADUtils
from bloodhound.ad.trusts import ADDomainTrust
from bloodhound.enumeration.memberships import MembershipEnumerator
from bloodhound.enumeration.acls import parse_binary_acl
from bloodhound.ad.structures import LDAP_SID
from frozendict import frozendict
from bloodhound.enumeration.outputworker import OutputWorker

from certipy.lib.constants import *
from certipy.lib.security import ActiveDirectorySecurity, CertificateSecurity, CASecurity
from certipy.commands.find import filetime_to_str
from asn1crypto import x509


class BloodHoundOutput:
    """Handles BloodHound-specific output processing."""
    
    def __init__(self, snapshot, output_folder, console: Console, 
                 sidcache, dncache, computersidcache, domains, objecttype_guid_map,
                 domaincontrollers, rootdomain, certtemplates):
        self.snap = snapshot
        self.output = output_folder
        self.console = console
        self.sidcache = sidcache
        self.dncache = dncache
        self.computersidcache = computersidcache
        self.domains = domains
        self.objecttype_guid_map = objecttype_guid_map
        self.domaincontrollers = domaincontrollers
        self.rootdomain = rootdomain
        self.certtemplates = certtemplates
        
        # Initialize counters
        self.numUsers = 0
        self.numGroups = 0
        self.numComputers = 0
        self.numTrusts = 0
        self.numCertTemplates = 0
        self.numCAs = 0
        self.trusts = []
        self.writeQueues = {}
        
        # Domain info (set in process())
        self.domainname = None
        self.domain_object = None
        self.domainsid = None

    def process(self):
        """Process all objects and generate BloodHound output."""
        self.domainname = ADUtils.ldap2domain(self.rootdomain)
        self.domain_object = self.snap.getObject(self.dncache[self.rootdomain])
        self.domainsid = ADUtils.get_entry_property(self.domain_object, 'objectSid')

        for ptype in ['users', 'computers', 'groups', 'domains', 'cert_bh', 'cert_ly4k_tpls', 'cert_ly4k_cas']:
            self.writeQueues[ptype] = queue.Queue()
            btype = ptype

            if ptype.startswith("cert_"):
                if ptype.endswith("bh"):
                    btype = "gpos"
                elif ptype.endswith("ly4k_tpls"):
                    btype = "templates"
                elif ptype.endswith("ly4k_cas"):
                    btype = "cas"
            
            results_worker = threading.Thread(
                target=OutputWorker.membership_write_worker,
                args=(self.writeQueues[ptype], btype, os.path.join(
                    self.output, f"{self.snap.header.server}_{self.snap.header.filetimeUnix}_{ptype}.json"
                ))
            )
            results_worker.daemon = True
            results_worker.start()

        for idx, obj in track(enumerate(self.snap.objects), description="Processing", total=self.snap.header.numObjects):
            for fun in [self.processUsers, self.processComputers, self.processGroups, 
                       self.processTrusts, self.processCertTemplates, self.processCAs]:
                ret = fun(obj)
                if ret:
                    break

        self.write_default_users()
        self.write_default_groups()
        self.processDomains()

        for ptype in ['users', 'computers', 'groups', 'domains', 'cert_bh', 'cert_ly4k_tpls', 'cert_ly4k_cas']:
            self.writeQueues[ptype].put(None)
            self.writeQueues[ptype].join()

        self.console.print(f"[green]✓[/green] Output written to {self.snap.header.server}_{self.snap.header.filetimeUnix}_*.json files")

    def processDomains(self):
        level_id = ADUtils.get_entry_property(self.domain_object, 'msds-behavior-version', -1)
        try:
            functional_level = ADUtils.FUNCTIONAL_LEVELS[int(level_id)]
        except KeyError:
            functional_level = 'Unknown'

        domain = {
            "ObjectIdentifier": ADUtils.get_entry_property(self.domain_object, 'objectSid'),
            "Properties": {
                "name": self.domainname.upper(),
                "domain": self.domainname.upper(),
                "domainsid": ADUtils.get_entry_property(self.domain_object, 'objectSid'),
                "distinguishedname": ADUtils.get_entry_property(self.domain_object, 'distinguishedName'),
                "description": ADUtils.get_entry_property(self.domain_object, 'description', ''),
                "functionallevel": functional_level,
                "Machine Account Quota": ADUtils.get_entry_property(self.domain_object, 'ms-DS-MachineAccountQuota'),
                "highvalue": True,
                "isaclprotected": False,
                "collected": True,
                "whencreated": ADUtils.get_entry_property(self.domain_object, 'whencreated', default=0)
            },
            "Trusts": [],
            "Aces": [],
            "Links": [],
            "ChildObjects": [],
            "GPOChanges": {
                "AffectedComputers": [],
                "DcomUsers": [],
                "LocalAdmins": [],
                "PSRemoteUsers": [],
                "RemoteDesktopUsers": []
            },
            "IsDeleted": False,
            "IsACLProtected": False
        }

        aces = self.parse_acl(domain, 'domain', ADUtils.get_entry_property(self.domain_object, 'nTSecurityDescriptor', raw=True))
        domain['Aces'] = self.resolve_aces(aces)
        domain['Trusts'] = self.trusts

        self.writeQueues["domains"].put(domain)

    def processComputers(self, entry):
        if not ADUtils.get_entry_property(entry, 'sAMAccountType', -1) == 805306369:
            return

        hostname = ADUtils.get_entry_property(entry, 'dNSHostName')
        if not hostname:
            resolved_entry = ADUtils.resolve_ad_entry(entry)
            hostname = resolved_entry['principal']

        distinguishedName = ADUtils.get_entry_property(entry, 'distinguishedName')

        membership_entry = {
            "attributes": {
                "objectSid": ADUtils.get_entry_property(entry, 'objectSid'),
                "primaryGroupID": ADUtils.get_entry_property(entry, 'primaryGroupID')
            }
        }

        computer = {
            'ObjectIdentifier': ADUtils.get_entry_property(entry, 'objectsid'),
            'AllowedToAct': [],
            'PrimaryGroupSID': MembershipEnumerator.get_primary_membership(membership_entry),
            'ContainedBy': None,
            'DumpSMSAPassword': [],
            'Properties': {
                'name': hostname.upper(),
                'domainsid': self.domainsid,
                'domain': self.domainname.upper(),
                'highvalue': False,
                'distinguishedname': distinguishedName
            },
            'LocalGroups': [],
            'LocalAdmins': {'Collected': False, 'FailureReason': None, 'Results': []},
            'RemoteDesktopUsers': {'Collected': False, 'FailureReason': None, 'Results': []},
            'DcomUsers': {'Collected': False, 'FailureReason': None, 'Results': []},
            'PSRemoteUsers': {'Collected': False, 'FailureReason': None, 'Results': []},
            'UserRights': [],
            'PrivilegedSessions': {
                'Collected': False,
                'FailureReason': None,
                'Results': []
            },
            'Sessions': {
                'Collected': False,
                'FailureReason': None,
                'Results': []
            },
            'RegistrySessions': {
                'Collected': False,
                'FailureReason': None,
                'Results': []
            },
            'AllowedToDelegate': [],
            'Aces': [],
            'HasSIDHistory': [],
            'IsDeleted': ADUtils.get_entry_property(entry, 'isDeleted', default=False),
            'Status': None
        }

        props = computer['Properties']
        props['unconstraineddelegation'] = ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 0x00080000 == 0x00080000
        props['enabled'] = ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 2 == 0
        props['trustedtoauth'] = ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 0x01000000 == 0x01000000
        props['samaccountname'] = ADUtils.get_entry_property(entry, 'sAMAccountName')
        props['haslaps'] = (
            ADUtils.get_entry_property(entry, 'ms-mcs-admpwdexpirationtime', 0) != 0
            or
            ADUtils.get_entry_property(entry, 'msLAPS-PasswordExpirationTime', 0) != 0
        )
        props['lastlogon'] = ADUtils.win_timestamp_to_unix(ADUtils.get_entry_property(entry, 'lastlogon', default=-1, raw=True))
        props['lastlogontimestamp'] = ADUtils.win_timestamp_to_unix(ADUtils.get_entry_property(entry, 'lastlogontimestamp', default=-1, raw=True))
        if props['lastlogontimestamp'] == 0:
            props['lastlogontimestamp'] = -1
        props['pwdlastset'] = ADUtils.win_timestamp_to_unix(ADUtils.get_entry_property(entry, 'pwdLastSet', default=-1, raw=True))
        props['whencreated'] = ADUtils.get_entry_property(entry, 'whencreated', default=0)
        props['serviceprincipalnames'] = ADUtils.get_entry_property(entry, 'servicePrincipalName', [])
        props['description'] = ADUtils.get_entry_property(entry, 'description')
        props['operatingsystem'] = ADUtils.get_entry_property(entry, 'operatingSystem')
        servicepack = ADUtils.get_entry_property(entry, 'operatingSystemServicePack')
        if servicepack:
            props['operatingsystem'] = '%s %s' % (props['operatingsystem'], servicepack)
        props['sidhistory'] = [LDAP_SID(bsid).formatCanonical() for bsid in ADUtils.get_entry_property(entry, 'sIDHistory', [])]

        delegatehosts = ADUtils.get_entry_property(entry, 'msDS-AllowedToDelegateTo', [])
        for host in delegatehosts:
            try:
                target = host.split('/')[1]
            except IndexError:
                logging.warning('Invalid delegation target: %s', host)
                continue
            try:
                sid = self.computersidcache[target]
                delegateObj = {
                    "ObjectIdentifier": sid,
                    "ObjectType": self.resolve_sid(sid)['ObjectType']
                }
                computer['AllowedToDelegate'].append(delegateObj)
            except KeyError:
                if '.' in target:
                    logging.warning('Unable to find sid for delegation target: %s', host)
                pass

        aces = self.parse_acl(computer, 'computer', ADUtils.get_entry_property(entry, 'msDS-AllowedToActOnBehalfOfOtherIdentity', raw=True))
        outdata = self.resolve_aces(aces)
        for delegated in outdata:
            if delegated['RightName'] == 'Owner':
                continue
            if delegated['RightName'] == 'GenericAll':
                computer['AllowedToAct'].append({'ObjectIdentifier': delegated['PrincipalSID'], 'ObjectType': delegated['PrincipalType']})

        aces = self.parse_acl(computer, 'computer', ADUtils.get_entry_property(entry, 'nTSecurityDescriptor', raw=True))
        computer['Aces'] = self.resolve_aces(aces)

        self.numComputers += 1
        self.writeQueues["computers"].put(computer)
        return True

    def processCertTemplates(self, entry):
        if not 'pkicertificatetemplate' in entry.classes:
            return

        name = ADUtils.get_entry_property(entry, 'name')
        if not name:
            return

        enabled = name in self.certtemplates
        object_identifier = ADUtils.get_entry_property(entry, 'objectGUID')
        validity_period = filetime_to_str(ADUtils.get_entry_property(entry, 'pKIExpirationPeriod'))
        renewal_period = filetime_to_str(ADUtils.get_entry_property(entry, 'pKIOverlapPeriod'))
        certificate_name_flag = ADUtils.get_entry_property(entry, 'msPKI-Certificate-Name-Flag', 0)
        certificate_name_flag = CertificateNameFlag(int(certificate_name_flag))
        enrollment_flag = ADUtils.get_entry_property(entry, 'msPKI-Enrollment-Flag', 0)
        enrollment_flag = EnrollmentFlag(int(enrollment_flag))
        authorized_signatures_required = int(ADUtils.get_entry_property(entry, 'msPKI-RA-Signature', 0))
        application_policies = ADUtils.get_entry_property(entry, 'msPKI-RA-Application-Policies', raw=True, default=[])
        application_policies = list(map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, application_policies))
        extended_key_usage = ADUtils.get_entry_property(entry, "pKIExtendedKeyUsage", default=[])
        extended_key_usage = list(map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, extended_key_usage))
        client_authentication = (any(eku in extended_key_usage for eku in ["Client Authentication", "Smart Card Logon", "PKINIT Client Authentication", "Any Purpose"]) or len(extended_key_usage) == 0)
        enrollment_agent = (any(eku in extended_key_usage for eku in ["Certificate Request Agent", "Any Purpose"]) or len(extended_key_usage) == 0)
        enrollee_supplies_subject = any(flag in certificate_name_flag for flag in [CertificateNameFlag.ENROLLEE_SUPPLIES_SUBJECT])
        requires_manager_approval = (EnrollmentFlag.PEND_ALL_REQUESTS in enrollment_flag)
        security = CertificateSecurity(ADUtils.get_entry_property(entry, "nTSecurityDescriptor", raw=True))
        aces = self.security_to_bloodhound_aces(security)

        certtemplate = {
            'Properties': {
                'highvalue': (enabled and any([all([enrollee_supplies_subject, not requires_manager_approval, client_authentication]), all([enrollment_agent, not requires_manager_approval])])),
                'name': "%s@%s" % (ADUtils.get_entry_property(entry, "CN").upper(), self.domainname.upper()),
                'type': 'Certificate Template',
                'domain': self.domainname.upper(),
                'Template Name': ADUtils.get_entry_property(entry, 'CN'),
                'Display Name': ADUtils.get_entry_property(entry, 'displayName'),
                'Client Authentication': client_authentication,
                'Enrollee Supplies Subject': enrollee_supplies_subject,
                'Extended Key Usage': extended_key_usage,
                'Requires Manager Approval': requires_manager_approval,
                'Validity Period': validity_period,
                'Renewal Period': renewal_period,
                'Certificate Name Flag': certificate_name_flag.to_str_list(),
                'Enrollment Flag': enrollment_flag.to_str_list(),
                'Authorized Signatures Required': authorized_signatures_required,
                'Application Policies': application_policies,
                'Enabled': enabled,
                'Certificate Authorities': list(self.certtemplates[name]),
            },
            'ObjectIdentifier': object_identifier.lstrip("{").rstrip("}"),
            'Aces': aces,
        }

        self.numCertTemplates += 1
        self.writeQueues["cert_bh"].put(certtemplate)
        self.writeQueues["cert_ly4k_tpls"].put(certtemplate)
        return True

    def processCAs(self, entry):
        if not 'pkienrollmentservice' in entry.classes:
            return
        
        name = ADUtils.get_entry_property(entry, 'name')
        if not name:
            return
        
        object_identifier = ADUtils.get_entry_property(entry, 'objectGUID')
        ca_name = ADUtils.get_entry_property(entry, 'cn')
        dns_name = ADUtils.get_entry_property(entry, 'dNSHostName')
        subject_name = ADUtils.get_entry_property(entry, 'cACertificateDN')
        ca_certificate = x509.Certificate.load(ADUtils.get_entry_property(entry, 'cACertificate', raw=True))["tbs_certificate"]
        serial_number = hex(int(ca_certificate["serial_number"]))[2:].upper()
        validity = ca_certificate["validity"].native
        validity_start = str(validity["not_before"])
        validity_end = str(validity["not_after"])
        security = CASecurity(ADUtils.get_entry_property(entry, "nTSecurityDescriptor"))
        aces = self.ca_security_to_bloodhound_aces(security)

        cas = {
            "Properties": {
                "highvalue": True,
                "name": "%s@%s" % (name.upper(), self.domainname.upper()),
                "domain": self.domainname.upper(),
                "type": "Enrollment Service",
                "CA Name": ca_name,
                "DNS Name": dns_name,
                "Certificate Subject": subject_name,
                "Certificate Serial Number": serial_number,
                "Certificate Validity Start": validity_start,
                "Certificate Validity End": validity_end,
                "Web Enrollment": "",
                "User Specified SAN": "",
                "Request Disposition": "",
            },
            "ObjectIdentifier": object_identifier.lstrip("{").rstrip("}"),
            "Aces": aces,
        }

        self.numCAs += 1
        self.writeQueues["cert_bh"].put(cas)
        self.writeQueues["cert_ly4k_cas"].put(cas)
        return True

    def processTrusts(self, entry):
        if 'trusteddomain' not in entry.classes:
            return

        domtrust = ADDomainTrust(
            ADUtils.get_entry_property(entry, 'name'),
            ADUtils.get_entry_property(entry, 'trustDirection'),
            ADUtils.get_entry_property(entry, 'trustType'),
            ADUtils.get_entry_property(entry, 'trustAttributes'),
            ADUtils.get_entry_property(entry, 'securityIdentifier', raw=True)
        )
        
        trust = domtrust.to_output()
        self.numTrusts += 1
        self.trusts.append(trust)
        return True

    def processGroups(self, entry):
        if not 'group' in entry.classes:
            return

        highvalue = ["S-1-5-32-544", "S-1-5-32-550", "S-1-5-32-549", "S-1-5-32-551", "S-1-5-32-548"]

        def is_highvalue(sid):
            if sid.endswith("-512") or sid.endswith("-516") or sid.endswith("-519") or sid.endswith("-520"):
                return True
            if sid in highvalue:
                return True
            return False

        distinguishedName = ADUtils.get_entry_property(entry, 'distinguishedName')
        resolved_entry = ADUtils.resolve_ad_entry(entry)
        sid = ADUtils.get_entry_property(entry, "objectSid")

        group = {
            "ObjectIdentifier": sid,
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "highvalue": is_highvalue(sid),
                "name": resolved_entry['principal'],
                "distinguishedname": distinguishedName,
            },
            "Members": [],
            "Aces": [],
            "IsDeleted": ADUtils.get_entry_property(entry, 'isDeleted', default=False),
            "IsACLProtected": False,
        }
        if sid in ADUtils.WELLKNOWN_SIDS:
            group['ObjectIdentifier'] = '%s-%s' % (self.domainname.upper(), sid)

        group['Properties']['admincount'] = ADUtils.get_entry_property(entry, 'adminCount', default=0) == 1
        group['Properties']['description'] = ADUtils.get_entry_property(entry, 'description', '')
        group['Properties']['whencreated'] = ADUtils.get_entry_property(entry, 'whencreated', default=0)
        group['Properties']['samaccountname'] = ADUtils.get_entry_property(entry, 'samaccountname')

        for member in ADUtils.get_entry_property(entry, 'member', []):
            resolved_member = self.get_membership(member)
            if resolved_member:
                group['Members'].append(resolved_member)

        aces = self.parse_acl(group, 'group', ADUtils.get_entry_property(entry, 'nTSecurityDescriptor', raw=True))
        group['Aces'] += self.resolve_aces(aces)

        self.numGroups += 1
        self.writeQueues["groups"].put(group)
        return True

    def processUsers(self, entry):
        if not (('user' in entry.classes and 'person' == entry.category) or 'msds-groupmanagedserviceaccount' in entry.classes):
            return

        distinguishedName = ADUtils.get_entry_property(entry, 'distinguishedName')
        resolved_entry = ADUtils.resolve_ad_entry(entry)
        if resolved_entry['type'] == 'trustaccount':
            return

        domain = ADUtils.ldap2domain(distinguishedName)

        membership_entry = {
            "attributes": {
                "objectSid": ADUtils.get_entry_property(entry, 'objectSid'),
                "primaryGroupID": ADUtils.get_entry_property(entry, 'primaryGroupID')
            }
        }

        user = {
            "AllowedToDelegate": [],
            "ObjectIdentifier": ADUtils.get_entry_property(entry, 'objectSid'),
            "PrimaryGroupSID": MembershipEnumerator.get_primary_membership(membership_entry),
            "ContainedBy": None,
            "Properties": {
                "name": resolved_entry['principal'],
                "domain": domain.upper(),
                "domainsid": self.domainsid,
                "highvalue": False,
                "distinguishedname": distinguishedName,
                "unconstraineddelegation": ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 0x00080000 == 0x00080000,
                "trustedtoauth": ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 0x01000000 == 0x01000000,
                "passwordnotreqd": ADUtils.get_entry_property(entry, 'userAccountControl', default=0) & 0x00000020 == 0x00000020
            },
            "Aces": [],
            "SPNTargets": [],
            "HasSIDHistory": [],
            "IsDeleted": ADUtils.get_entry_property(entry, 'isDeleted', default=False),
            "IsACLProtected": False,
        }

        MembershipEnumerator.add_user_properties(user, entry)

        if 'allowedtodelegate' in user['Properties']:
            for host in user['Properties']['allowedtodelegate']:
                try:
                    target = host.split('/')[1]
                except IndexError:
                    logging.warning('Invalid delegation target: %s', host)
                    continue
                try:
                    sid = self.computersidcache[target]
                    delegateObj = {
                        "ObjectIdentifier": sid,
                        "ObjectType": self.resolve_sid(sid)['ObjectType']
                    }
                    user['AllowedToDelegate'].append(delegateObj)
                except KeyError:
                    logging.warning('Unable to find sid for delegation target: %s', host)
                    pass
            del user['Properties']['allowedtodelegate']
        if len(user['Properties']['sidhistory']) > 0:
            for historysid in user['Properties']['sidhistory']:
                user['HasSIDHistory'].append(self.resolve_sid(historysid))

        if ADUtils.get_entry_property(entry, 'msDS-GroupMSAMembership', default=b'', raw=True) != b'':
            aces = self.parse_acl(user, 'user', ADUtils.get_entry_property(entry, 'msDS-GroupMSAMembership', raw=True))
            processed_aces = self.resolve_aces(aces)
            for ace in processed_aces:
                if ace['RightName'] == 'Owner':
                    continue
                ace['RightName'] = 'ReadGMSAPassword'
                user['Aces'].append(ace)

        aces = self.parse_acl(user, 'user', ADUtils.get_entry_property(entry, 'nTSecurityDescriptor', raw=True))
        user['Aces'] += self.resolve_aces(aces)

        self.numUsers += 1
        self.writeQueues["users"].put(user)
        return True

    @functools.lru_cache(maxsize=4096)
    def resolve_aces(self, aces):
        aces_out = []
        for ace in aces:
            out = {
                'RightName': ace['rightname'],
                'IsInherited': ace['inherited']
            }
            if ace['sid'] in ADUtils.WELLKNOWN_SIDS:
                out['PrincipalSID'] = u'%s-%s' % (self.domainname.upper(), ace['sid'])
                out['PrincipalType'] = ADUtils.WELLKNOWN_SIDS[ace['sid']][1].capitalize()
            else:
                try:
                    entry = self.snap.getObject(self.sidcache[ace['sid']])
                except KeyError:
                    entry = {'type': 'Unknown', 'principal': ace['sid']}
                resolved_entry = ADUtils.resolve_ad_entry(entry)
                out['PrincipalSID'] = ace['sid']
                out['PrincipalType'] = resolved_entry['type']
            aces_out.append(out)
        return aces_out

    @functools.lru_cache(maxsize=4096)
    def _parse_acl_cached(self, parselaps, entrytype, acl):
        fake_entry = {"Properties": {"haslaps": True if parselaps else False}}
        _, aces = parse_binary_acl(fake_entry, entrytype, acl, self.objecttype_guid_map)
        for i, ace in enumerate(aces):
            aces[i] = frozendict(ace)
        return frozenset(aces)

    def parse_acl(self, entry, entrytype, acl):
        parselaps = entrytype == 'computer' and entry['Properties']['haslaps'] and (
            "ms-mcs-admpwd" in self.objecttype_guid_map or
            "ms-laps-password" in self.objecttype_guid_map or
            "ms-laps-encryptedpassword" in self.objecttype_guid_map
        )
        aces = self._parse_acl_cached(parselaps, entrytype, acl)
        self.cacheInfo = self._parse_acl_cached.cache_info()
        return aces

    @functools.lru_cache(maxsize=2048)
    def resolve_sid(self, sid):
        out = {}
        if sid in ADUtils.WELLKNOWN_SIDS:
            out['ObjectID'] = u'%s-%s' % (self.domainname.upper(), sid)
            out['ObjectType'] = ADUtils.WELLKNOWN_SIDS[sid][1].capitalize()
        else:
            try:
                entry = self.snap.getObject(self.sidcache[sid])
            except KeyError:
                entry = {'type': 'Unknown', 'principal': sid}
            resolved_entry = ADUtils.resolve_ad_entry(entry)
            out['ObjectID'] = sid
            out['ObjectType'] = resolved_entry['type']
        return out

    @functools.lru_cache(maxsize=2048)
    def get_membership(self, member):
        try:
            entry = self.snap.getObject(self.dncache[member])
        except KeyError:
            return None
        resolved_entry = ADUtils.resolve_ad_entry(entry)
        return {
            "ObjectIdentifier": resolved_entry['objectid'],
            "ObjectType": resolved_entry['type'].capitalize()
        }

    def write_default_users(self):
        user = {
            "AllowedToDelegate": [],
            "ObjectIdentifier": "%s-S-1-5-20" % self.domainname.upper(),
            "PrimaryGroupSID": None,
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "name": "NT AUTHORITY@%s" % self.domainname.upper(),
                "highvalue": False,
            },
            "Aces": [],
            "SPNTargets": [],
            "HasSIDHistory": [],
            "IsDeleted": False,
            "ContainedBy": None,
            "IsACLProtected": False,
        }
        self.writeQueues["users"].put(user)

    def write_default_groups(self):
        group = {
            "ObjectIdentifier": "%s-S-1-5-9" % self.domainname.upper(),
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "name": "ENTERPRISE DOMAIN CONTROLLERS@%s" % self.domainname.upper()
            },
            "ContainedBy": None,
            "Members": [],
            "Aces": [],
            "IsDeleted": False,
            "IsACLProtected": False,
            "ContainedBy": None
        }

        for dc in self.domaincontrollers:
            entry = self.snap.getObject(dc)
            resolved_entry = ADUtils.resolve_ad_entry(entry)
            memberdata = {
                "ObjectIdentifier": resolved_entry['objectid'],
                "ObjectType": resolved_entry['type'].capitalize()
            }
            group["Members"].append(memberdata)

        self.writeQueues["groups"].put(group)

        evgroup = {
            "ObjectIdentifier": "%s-S-1-1-0" % self.domainname.upper(),
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "name": "EVERYONE@%s" % self.domainname.upper()
            },
            "Members": [],
            "Aces": [],
            "IsDeleted": False,
            "IsACLProtected": False,
            "ContainedBy": None
        }
        self.writeQueues["groups"].put(evgroup)

        augroup = {
            "ObjectIdentifier": "%s-S-1-5-11" % self.domainname.upper(),
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "name": "AUTHENTICATED USERS@%s" % self.domainname.upper()
            },
            "Members": [],
            "Aces": [],
            "IsDeleted": False,
            "IsACLProtected": False,
            "ContainedBy": None
        }
        self.writeQueues["groups"].put(augroup)

        iugroup = {
            "ObjectIdentifier": "%s-S-1-5-4" % self.domainname.upper(),
            "Properties": {
                "domain": self.domainname.upper(),
                "domainsid": self.domainsid,
                "name": "INTERACTIVE@%s" % self.domainname.upper()
            },
            "Members": [],
            "Aces": [],
            "IsDeleted": False,
            "IsACLProtected": False,
            "ContainedBy": None
        }
        self.writeQueues["groups"].put(iugroup)

    def security_to_bloodhound_aces(self, security: ActiveDirectorySecurity) -> List:
        aces = []
        principal_type = ""

        owner_sid = security.owner
        if owner_sid in ADUtils.WELLKNOWN_SIDS:
            principal = u'%s-%s' % (self.domainname.upper(), owner_sid)
            principal_type = ADUtils.WELLKNOWN_SIDS[owner_sid][1].capitalize()
        else:
            try:
                entry = self.snap.getObject(self.sidcache[owner_sid])
                resolved_entry = ADUtils.resolve_ad_entry(entry)
                principal_type = resolved_entry['type']
            except KeyError:
                entry = {'type': 'Unknown', 'principal': owner_sid}
        aces.append({
            "PrincipalSID": owner_sid,
            "PrincipalType": principal_type,
            "RightName": "Owner",
            "IsInherited": False,
        })

        for sid, rights in security.aces.items():
            principal = sid
            principal_type = ""

            if sid in ADUtils.WELLKNOWN_SIDS:
                principal = u'%s-%s' % (self.domainname.upper(), sid)
                principal_type = ADUtils.WELLKNOWN_SIDS[sid][1].capitalize()
            else:
                try:
                    entry = self.snap.getObject(self.sidcache[sid])
                    resolved_entry = ADUtils.resolve_ad_entry(entry)
                    principal_type = resolved_entry['type']
                except KeyError:
                    entry = {'type': 'Unknown', 'principal': sid}

            try:
                standard_rights = list(rights["rights"])
            except:
                standard_rights = rights["rights"].to_list()

            for right in standard_rights:
                aces.append({
                    "PrincipalSID": principal,
                    "PrincipalType": principal_type,
                    "RightName": str(right),
                    "IsInherited": False,
                })

            extended_rights = rights["extended_rights"]
            for extended_right in extended_rights:
                aces.append({
                    "PrincipalSID": principal,
                    "PrincipalType": principal_type,
                    "RightName": EXTENDED_RIGHTS_MAP[extended_right].replace("-", "") if extended_right in EXTENDED_RIGHTS_MAP else extended_right,
                    "IsInherited": False,
                })

        return aces

    def ca_security_to_bloodhound_aces(self, security: ActiveDirectorySecurity) -> List:
        aces = []
        principal_type = ""

        for sid, rights in security.aces.items():
            principal = sid
            principal_type = ""

            if sid in ADUtils.WELLKNOWN_SIDS:
                principal = u'%s-%s' % (self.domainname.upper(), sid)
                principal_type = ADUtils.WELLKNOWN_SIDS[sid][1].capitalize()
            else:
                try:
                    entry = self.snap.getObject(self.sidcache[sid])
                    resolved_entry = ADUtils.resolve_ad_entry(entry)
                    principal_type = resolved_entry['type']
                except KeyError:
                    entry = {'type': 'Unknown', 'principal': sid}

            try:
                standard_rights = list(rights["rights"])
            except:
                standard_rights = rights["rights"].to_list()

            for right in standard_rights:
                if not principal_type == "Computer":
                    aces.append({
                        "PrincipalSID": principal,
                        "PrincipalType": principal_type,
                        "RightName": str(right),
                        "IsInherited": False,
                    })

            extended_rights = rights["extended_rights"]
            for extended_right in extended_rights:
                aces.append({
                    "PrincipalSID": principal,
                    "PrincipalType": principal_type,
                    "RightName": EXTENDED_RIGHTS_MAP[extended_right].replace("-", "") if extended_right in EXTENDED_RIGHTS_MAP else extended_right,
                    "IsInherited": False,
                })

        return aces




"""
Builds the treeview structure for a naming context.
"""

from collections import defaultdict
from io import BytesIO
from adexpsnapshot.treeview.structure import treeview_structure

def build_nc_tree(snap, nc_root, nc_filter, dncache=None, synthetic_collector=None):
    """
    Build tree for a naming context.
    
    Args:
        snap: Snapshot object
        nc_root: Root DN of this NC
        nc_filter: Filter function for DNs
        dncache: DN to index cache
        synthetic_collector: Optional dict to collect synthetic DN -> True for missing parents
    
    Returns:
        Tree dict or None
    """
    dn_to_info = {}
    
    # Add real objects from cache
    for dn, obj_idx in dncache.items():
        if nc_filter(dn):
            dn_to_info[dn] = {
                'obj_idx': obj_idx,
                'obj_offset': snap.objectOffsets[obj_idx],
                'dn': dn
            }
    
    if not dn_to_info:
        return None  # No objects found in this NC
    
    children_map = defaultdict(list)
    
    def get_parent_dn(dn):
        r"""
        Extract parent DN, handling escaped commas.
        LDAP DNs use backslash escaping: CN=Smith\, John,OU=Users
        """
        # Find first unescaped comma
        i = 0
        while i < len(dn):
            if dn[i] == ',' and (i == 0 or dn[i-1] != '\\'):
                # Found unescaped comma - parent is everything after
                return dn[i+1:] if i+1 < len(dn) else None
            i += 1
        return None  # No parent (root)
    
    # Build parent-child relationships, creating synthetics on-demand
    # Use negative indices as temporary IDs for synthetic nodes
    next_synthetic_idx = -1
    newly_created = []  # Track synthetics to process after main loop
    
    for dn in list(dn_to_info.keys()):
        if dn == nc_root:
            continue
        
        parent = get_parent_dn(dn)
        if parent:
            # Create synthetic entries for all missing ancestors
            current = parent
            while current and current not in dn_to_info:
                if nc_filter(current):
                    # Mark for synthetic creation
                    if synthetic_collector is not None:
                        synthetic_collector[current] = True
                    # Create placeholder entry with temporary negative index
                    temp_idx = next_synthetic_idx
                    dn_to_info[current] = {
                        'obj_idx': temp_idx,
                        'obj_offset': 0,  # Placeholder, will be updated
                        'dn': current
                    }
                    newly_created.append(current)
                    next_synthetic_idx -= 1
                current = get_parent_dn(current)
                if current == nc_root:
                    break
            
            # Now add to children map
            if parent in dn_to_info:
                children_map[parent].append(dn)
    
    # Process newly created synthetic DNs to add them to their parents' children_map
    for dn in newly_created:
        parent = get_parent_dn(dn)
        if parent and parent in dn_to_info:
            children_map[parent].append(dn)
    
    root_dn = min(dn_to_info.keys(), key=len)
    
    def build_node(dn):
        info = dn_to_info[dn]
        child_dns = children_map.get(dn, [])
        return {
            **info,
            'children': [build_node(c) for c in child_dns]
        }
    
    return build_node(root_dn)


def build_stub_nc_tree(nc_root, synthetic_collector=None):
    """
    Build a minimal single-node NC tree when a naming context has no LDAP objects.

    Bloodhound-style snapshots omit Schema NC enumeration but ADExplorer still
    expects a Schema section in treeview metadata.
    """
    if synthetic_collector is not None:
        synthetic_collector[nc_root] = True

    return {
        'obj_idx': -1,
        'obj_offset': 0,
        'dn': nc_root,
        'children': [],
    }


def encode_section(tree):
    """
    Universal section encoder using cstruct for type safety.
    
    Works identically for Domain, Configuration, and Schema NCs.
    No special cases, pure algorithm based on tree structure.
    
    Optimization: Leaf children are stored ONLY inline in their parent,
    not as standalone entries (eliminates redundancy).
    """
    
    # Flatten tree to get all nodes
    def flatten(node):
        result = [node]
        for c in node.get('children', []):
            result.extend(flatten(c))
        return result
    
    all_nodes = flatten(tree)
    
    # PASS 1: Identify which nodes need standalone entries
    # Leaf children are stored inline in parent - they don't need standalone entries
    inline_only = set()
    
    def mark_inline_leaves(node):
        """Mark leaf children that will be stored inline only."""
        children = node.get('children', [])
        for child in children:
            if not child.get('children'):
                # This is a leaf child - will be stored inline in parent
                inline_only.add(child['obj_idx'])
            else:
                # Recursively check this child's children
                mark_inline_leaves(child)
    
    mark_inline_leaves(tree)
    
    # PASS 2: Compute positions and cache children classification
    # Only allocate positions for: root, parents, and children-with-children
    node_word_positions = {}
    node_metadata = {}
    current_word_pos = 0
    
    for node in all_nodes:
        obj_idx = node['obj_idx']
        
        # Skip leaf children - they're stored inline in parent only
        if obj_idx in inline_only:
            continue
        
        node_children = node.get('children', [])
        
        # Position this node
        node_word_positions[obj_idx] = current_word_pos
        
        # Classify children ONCE
        entries_with_children = []
        entries_without_children = []
        for c in node_children:
            if c.get('children'):
                entries_with_children.append(c)
            else:
                entries_without_children.append(c)

        # ParentNode size: header(16) + child_offsets(count*4) + inline_child_offsets(count*8)
        size = 4 + len(entries_with_children) + len(entries_without_children) * 2

        # Cache for PASS 3 (just the lists, compute len() when needed)
        node_metadata[obj_idx] = (entries_with_children, entries_without_children)

        current_word_pos += size
    
    # PASS 3: Write data using cstruct structures
    output = BytesIO()
    output.write(b'\x00' * (current_word_pos * 4))  # Pre-allocate
    
    for node in all_nodes:
        obj_idx = node['obj_idx']
        
        # Skip inline-only nodes
        if obj_idx in inline_only:
            continue
        
        word_pos = node_word_positions[obj_idx]
        byte_pos = word_pos * 4
        
        output.seek(byte_pos)
        
        # Write ParentNode structure (all non-inline nodes are parents)
        entries_with_children, entries_without_children = node_metadata[obj_idx]
        
        parent = treeview_structure.ParentNode()
        parent.objectOffset = node['obj_offset']
        parent.num_children_with_children = len(entries_with_children)
        parent.num_children_without_children = len(entries_without_children)
        
        # Build child_offsets array (relative byte offsets to child ParentNode entries)
        parent.child_offsets = []
        for child in entries_with_children:
            child_word_pos = node_word_positions[child['obj_idx']]
            relative_byte_offset = (child_word_pos - word_pos) * 4
            parent.child_offsets.append(relative_byte_offset)
        
        # Build inline_child_offsets array (direct objectOffsets for leaf children)
        parent.inline_child_offsets = []
        for child in entries_without_children:
            parent.inline_child_offsets.append(child['obj_offset'])
        
        output.write(parent.dumps())
    
    return output.getvalue()






from dissect.cstruct import cstruct

structure = cstruct()
structure.load("""
    struct Header {
        char winAdSig[10];
        int marker;

        uint64 filetime;
        wchar optionalDescription[260];
        wchar  server[260];

        uint32 numObjects;
        uint32 numAttributes;

        uint64 metadataOffset;
        uint64 treeviewOffset;

        int unk0x43a;
    };

    struct MappingEntry {
        uint32 attrIndex; // index of the attribute in the global attribute table
        int attrOffset; // offset to the value of the attribute in this object
    }

    struct Object {
        uint32 objSize;
        uint32 tableSize;
        MappingEntry mappingTable[tableSize];

        //char blob[objSize - 4 - 4 - (tableSize * 8)]; // omitted as we use file offsetting instead
    };

    struct Property {
        uint32 lenPropName;
        wchar propName[lenPropName/2];
        int unk1;
        uint32 adsType;
        uint32 lenDN;
        wchar DN[lenDN/2];
        char schemaIDGUID[16];
        char attributeSecurityGUID[16];
        char blob[4];
    };

    struct Properties {
        uint32 numProperties;
        Property properties[numProperties];
    };

    struct SystemPossSuperior {
        uint32 lenSystemPossSuperior;
        wchar systemPossSuperior[lenSystemPossSuperior/2];
    };

    struct AuxiliaryClasses {
        uint32 lenAuxiliaryClass;
        wchar auxiliaryClass[lenAuxiliaryClass/2];
    };

    struct Block {
        uint32 unk1;
        uint32 unk2;
        wchar unk3[unk2/2];
    };

    struct Class {
        uint32 lenClassName;
        wchar className[lenClassName/2];
        uint32 lenDN;
        wchar DN[lenDN/2];

        uint32 lenCommonClassName; // in AD Explorer, some class names are common and some are considered advanced - the common ones are given a description
        wchar commonClassName[lenCommonClassName/2];

        uint32 lenSubClassOf;
        wchar subClassOf[lenSubClassOf/2];

        char schemaIDGUID[16];

        uint32 offsetToNumBlocks;
        char unk2[offsetToNumBlocks];

        uint32 numBlocks;
        Block blocks[numBlocks];

        uint32 numExtraShiz;
        char extraShiz[numExtraShiz*0x10];

        uint32 numSystemPossSuperiors;
        SystemPossSuperior systemPossSuperiors[numSystemPossSuperiors];

        uint32 numAuxiliaryClasses;
        AuxiliaryClasses auxiliaryClasses[numAuxiliaryClasses];

    };

    struct Classes {
        uint32 numClasses;
        Class classes[numClasses];
    };

    struct Right {
        uint32 lenName;
        wchar name[lenName/2];
        uint32 lenDesc;
        wchar desc[lenDesc/2];
        char blob[20];
    };

    struct Rights {
        uint32 numRights;
        Right rights[numRights];
    };

    // https://docs.microsoft.com/en-us/windows/win32/api/minwinbase/ns-minwinbase-systemtime
    struct SystemTime {
        WORD wYear;
        WORD wMonth;
        WORD wDayOfWeek;
        WORD wDay;
        WORD wHour;
        WORD wMinute;
        WORD wSecond;
        WORD wMilliseconds;
    };

""", compiled=True)



"""Objects output mode - outputs all objects as NDJSON."""
import codecs
import json
import base64
import logging
import os
import queue
import threading
from rich.console import Console
from rich.progress import track


class ObjectsOutput:
    """Handles Objects-specific output processing."""
    
    def __init__(self, snapshot, output_folder, console: Console):
        self.snap = snapshot
        self.output = output_folder
        self.console = console
        self.outputfile = f"{snapshot.header.server}_{snapshot.header.filetimeUnix}_objects.ndjson"

    def process(self):
        """Process all objects and output to NDJSON file."""
        class BaseSafeEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, bytes):
                    return base64.b64encode(obj).decode("ascii")
                return json.JSONEncoder.default(self, obj)

        def write_worker(result_q, filename):
            try:
                fh_out = codecs.open(filename, 'w', 'utf-8')
            except:
                logging.warning('Could not write file: %s', filename)
                result_q.task_done()
                return

            wroteOnce = False
            while True:
                data = result_q.get()

                if data is None:
                    break

                if not wroteOnce:
                    wroteOnce = True
                else:
                    fh_out.write('\n')

                try:
                    encoded_member = json.dumps(data, indent=None, cls=BaseSafeEncoder)
                    fh_out.write(encoded_member)
                except TypeError:
                    logging.error('Data error {0}, could not convert data to json'.format(repr(data)))
                result_q.task_done()

            fh_out.close()
            result_q.task_done()
            
        wq = queue.Queue()
        results_worker = threading.Thread(target=write_worker, args=(wq, os.path.join(self.output, self.outputfile)))
        results_worker.daemon = True
        results_worker.start()

        for idx, obj in track(enumerate(self.snap.objects), description="Dumping objects", total=self.snap.header.numObjects):
            wq.put((dict(obj.attributes.data)))

        wq.put(None)
        wq.join()

        self.console.print(f"[green]✓[/green] Output written to {self.outputfile}")



"""
Treeview metadata enrichment functionality.

Detects if treeview metadata is missing and reconstructs it if needed.
"""

from enum import Enum
import logging
import shutil
import struct
from pathlib import Path
from adexpsnapshot.treeview.section_encoder import build_nc_tree, build_stub_nc_tree, encode_section
from adexpsnapshot.treeview.structure import treeview_structure
from adexpsnapshot.treeview.synthetic import create_synthetic_objects_data

TREEVIEW_MAGIC_ADES = b'\xFE\xFF\xFF\xFF\xFF\xFF\xFF\xFF'  # Likely ADES / populated
TREEVIEW_MAGIC_BOF  = b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'  # Likely BOF / unpopulated

class TreeviewStatus(Enum):
    INVALID = 0
    MISSING = 1
    UNPOPULATED = 2
    POPULATED = 3

def check_treeview_exists(snap):

    treeview_offset = snap.header.treeviewOffset
    if treeview_offset == 0:
        return TreeviewStatus.INVALID
    
    snap.fh.seek(treeview_offset)
    magic = snap.fh.read(8)
    
    if magic == TREEVIEW_MAGIC_ADES:
        header_data = snap.fh.read(20)
        if len(header_data) < 20:
            return TreeviewStatus.MISSING
        else:
            return TreeviewStatus.POPULATED

    if magic == TREEVIEW_MAGIC_BOF:
        return TreeviewStatus.UNPOPULATED

    return TreeviewStatus.INVALID

def get_output_filename(input_path):
    """
    Generate output filename: input.enriched.ext
    
    Examples:
    - snapshot.dat -> snapshot.enriched.dat
    - snapshot -> snapshot.enriched
    - file.xyz -> file.enriched.xyz
    """
    path = Path(input_path.name)
    
    if path.suffix:
        # Has extension: insert .enriched before it
        stem = path.stem
        suffix = path.suffix
        new_name = f"{stem}.enriched{suffix}"
    else:
        # No extension: just append .enriched
        new_name = f"{path.name}.enriched"
    
    return input_path.parent / new_name

def enrich_snapshot(ades):
    snap = ades.snap
    snapshot_path = Path(ades.snapfile.name)
        
    treeview_status = check_treeview_exists(snap)
    
    if treeview_status == TreeviewStatus.POPULATED:
        logging.info("Treeview metadata already exists, aborting.")
        return True
    elif treeview_status == TreeviewStatus.UNPOPULATED:
        logging.info("Treeview metadata is unpopulated, reconstructing...")
    elif treeview_status == TreeviewStatus.MISSING:
        logging.warning("Treeview metadata was expected but missing, reconstructing...")
    elif treeview_status == TreeviewStatus.INVALID:
        logging.error("Treeview metadata is invalid, cannot enrich snapshot")
        return False

    # Use existing preprocessing with cache
    ades.preprocessCached()
    dncache = ades.dncache
        
    # Identify NC roots from the snapshot
    if not ades.domain_dn or not ades.config_dn or not ades.schema_dn or not ades.forest_dn:
        logging.error("No domain, config, schema, or forest root found in snapshot!")
        return False
    
    domain_dns_zones_root = f"DC=DomainDnsZones,{ades.domain_dn}"
    forest_dns_zones_root = f"DC=ForestDnsZones,{ades.forest_dn}"

    logging.info(f"Domain root: {ades.domain_dn}")
    logging.info(f"Config root: {ades.config_dn}")
    logging.info(f"Schema root: {ades.schema_dn}")
    logging.info(f"Domain DNS Zones root: {domain_dns_zones_root}")
    logging.info(f"Forest DNS Zones root: {forest_dns_zones_root}")
    
    # Check which optional NCs exist
    has_domain_dns = any(dn.endswith(domain_dns_zones_root) for dn in dncache.keys())
    has_forest_dns = any(dn.endswith(forest_dns_zones_root) for dn in dncache.keys())
    
    # Build trees and collect missing parents on-demand
    logging.info("Building trees...")
    synthetic_collector = {}  # Will be populated during tree building
    
    domain_tree = build_nc_tree(snap, ades.domain_dn, lambda dn: dn.endswith(ades.domain_dn) and not dn.endswith(ades.config_dn) and not dn.endswith(domain_dns_zones_root) and not dn.endswith(forest_dns_zones_root), dncache, synthetic_collector)
    config_tree = build_nc_tree(snap, ades.config_dn, lambda dn: dn.endswith(ades.config_dn) and not dn.endswith(ades.schema_dn), dncache, synthetic_collector)
    schema_tree = build_nc_tree(snap, ades.schema_dn, lambda dn: dn.endswith(ades.schema_dn), dncache, synthetic_collector)
    if schema_tree is None:
        logging.info(
            "No Schema NC objects in snapshot; using synthetic schema root for treeview"
        )
        schema_tree = build_stub_nc_tree(ades.schema_dn, synthetic_collector)

    missing_required = [name for name, tree in (("Domain", domain_tree), ("Configuration", config_tree), ("Schema", schema_tree)) if tree is None]
    if missing_required:
        logging.error(f"Missing required NC tree(s): {', '.join(missing_required)}. Snapshot may be incomplete; cannot enrich.")
        return False

    domain_dns_zones_tree = None
    if has_domain_dns:
        domain_dns_zones_tree = build_nc_tree(snap, domain_dns_zones_root, lambda dn: dn.endswith(domain_dns_zones_root), dncache, synthetic_collector)

    forest_dns_zones_tree = None
    if has_forest_dns:
        forest_dns_zones_tree = build_nc_tree(snap, forest_dns_zones_root, lambda dn: dn.endswith(forest_dns_zones_root), dncache, synthetic_collector)
    
    # Build section names
    section_trees = [domain_tree, config_tree, schema_tree]
    if domain_dns_zones_tree:
        section_trees.append(domain_dns_zones_tree)
    if forest_dns_zones_tree:
        section_trees.append(forest_dns_zones_tree)

    # Create synthetic objects for all missing parents found during tree building
    synthetic_data = b''
    if synthetic_collector:
        logging.info(f"Creating {len(synthetic_collector)} synthetic containers...")
        
        # Synthetics go at treeview_offset
        synthetic_objects, synthetic_data = create_synthetic_objects_data(
            set(synthetic_collector.keys()), snap, snap.header.treeviewOffset
        )
        
        # Update all trees with actual synthetic object offsets
        def update_synthetics(node):
            """Recursively update synthetic nodes with real obj_offset."""
            dn = node['dn']
            if node['obj_idx'] < 0:
                node['obj_offset'] = synthetic_objects[dn]
            for child in node.get('children', []):
                update_synthetics(child)
        
        for tree in section_trees:
            update_synthetics(tree)
    
    # Encode all sections
    sections = []
    for tree in section_trees:
        section_data = bytes(encode_section(tree))
        sections.append(section_data)
    
    # Calculate header size
    tv_header_size = 16 + (len(section_trees) * 4)
    
    # Calculate section offsets (relative to treeview start, after header)
    # Layout: [Treeview Header][Section1][Section2][...]
    section_offsets = []
    current_offset = tv_header_size
    for section in sections:
        section_offsets.append(current_offset)
        current_offset += len(section)
    
    tv_header_struct = treeview_structure.TreeviewHeader()
    tv_header_struct.magic = 0xFFFFFFFFFFFFFFFE  # Combined 64-bit magic
    tv_header_struct.num_NCs = len(section_trees)  # Dynamic: 3-5 depending on DNS zones presence
    tv_header_struct.reserved = 0
    tv_header_struct.section_offsets = section_offsets
    
    tv_header = tv_header_struct.dumps()
    
    # Create enriched file by copying original snapshot first
    output_path = get_output_filename(snapshot_path)
    logging.info(f"Writing enriched snapshot to: {output_path}")
    try:
        shutil.copyfile(snapshot_path, output_path)
    except OSError as exc:
        logging.error(f"Failed to copy snapshot: {exc}")
        return False
    
    # Update treeviewOffset in main file header if we have synthetic objects
    # New treeview starts after synthetic objects
    old_treeview_offset = snap.header.treeviewOffset
    new_treeview_offset = snap.header.treeviewOffset + len(synthetic_data)
        
    with open(output_path, 'r+b') as out:
        if snap.header.winAdSig != b'win':
            logging.info("Fixing header signature")
            out.seek(0)
            out.write(b'win')
        
        if synthetic_data:
            out.seek(snap.header.fields['treeviewOffset'].offset)
            out.write(struct.pack("<Q", new_treeview_offset))

            out.seek(old_treeview_offset)
            out.write(synthetic_data)

            logging.info(f"Updated treeview offset: 0x{old_treeview_offset:x} → 0x{new_treeview_offset:x}")
        
        out.seek(new_treeview_offset)
        
        # Write treeview metadata: [Header][Sections...]
        out.write(tv_header)
        for section in sections:
            out.write(section)
        
        out.truncate(out.tell())
    
    return True